#include "php_orbit.h"
#include "phorb.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define CORBA_dup(t, v)	(t)CORBA_Object_duplicate((CORBA_Object)(v), NULL)

static zend_class_entry *corbatc_class;

void phorb_init_typecode_class(zend_class_entry *tcclass)
{
	corbatc_class = tcclass;
}

void phorb_fini_typecode_class(void)
{
}

gboolean phorb_is_typecode(zval *val)
{
	return Z_TYPE_P(val) == IS_OBJECT && Z_OBJCE_P(val) == corbatc_class;
}

zval *phorb_init_typecode(zval *res)
{
	object_init_ex(res, corbatc_class);

	return res;
}

ulong phorb_register_typecode(CORBA_TypeCode tc)
{
	return phorb_register_object((CORBA_Object)tc);
}

CORBA_TypeCode phorb_find_typecode(phorbIDL *idl, const char *repoid)
{
	GSList *list;
	CORBA_TypeCode ret = CORBA_OBJECT_NIL;
	OrbSLS_FETCH();

	if (idl)
		ret = g_hash_table_lookup(idl->typecodes, repoid);
	else {
		for(list = OrbSG(idls); list; list = list->next) {
			idl = (phorbIDL *)list->data;
			ret = g_hash_table_lookup(idl->typecodes, repoid);
			if (ret)
				break;
		}
	}

	return ret ? CORBA_dup(CORBA_TypeCode, ret) : CORBA_OBJECT_NIL;
}

void phorb_store_typecode(phorbIDL *idl, CORBA_TypeCode tc)
{
	tc = CORBA_dup(CORBA_TypeCode, tc);
	g_hash_table_insert(idl->typecodes, tc->repo_id, tc);
}

#define ADD_BASIC(type)	phorb_store_typecode(idl, TC_CORBA_##type)

void phorb_setup_typecodes(phorbIDL *idl)
{
	ADD_BASIC(short);
	ADD_BASIC(long);
	ADD_BASIC(long_long);
	ADD_BASIC(unsigned_short);
	ADD_BASIC(unsigned_long);
	ADD_BASIC(unsigned_long_long);
	ADD_BASIC(float);
	ADD_BASIC(double);
	ADD_BASIC(long_double);
	ADD_BASIC(boolean);
	ADD_BASIC(char);
	ADD_BASIC(wchar);
	ADD_BASIC(octet);
	ADD_BASIC(any);
	ADD_BASIC(TypeCode);
	ADD_BASIC(Object);
	ADD_BASIC(string);
	ADD_BASIC(wstring);
}

#undef ADD_BASIC

void phorb_invoke_typecode_method(INTERNAL_FUNCTION_PARAMETERS,
                                  zend_property_reference *props)
{
	zval *handle, *pobj = props->object;
	zend_overloaded_element	*fname =
		(zend_overloaded_element *)props->elements_list->tail->data;
	char *method;
	char *name;
	int name_len;
	CORBA_TypeCode tc;
	OrbSLS_FETCH();

	convert_to_string(&fname->element);
	method = Z_STRVAL(fname->element);
	if (!method) {
		zend_error(E_WARNING, "Can't make method name a string");
		goto error;
	}

	if (strcmp(method, "equal") == 0) {
		CORBA_Environment ev;
		CORBA_TypeCode tc2;
		zval *arg, **tmp;
		int type;

		if (zend_parse_parameters_ex(ZEND_PARSE_PARAMS_QUIET,
		                             ZEND_NUM_ARGS() TSRMLS_CC, "O",
		                             &arg, corbatc_class) == FAILURE) {
			zend_error(E_WARNING, "equal() expects exactly one "
			                      "CORBA_Typecode parameter");
			goto leave;
		}

		zend_hash_index_find(Z_OBJPROP_P(pobj), 0, (void *)&tmp);
		tc = (CORBA_TypeCode)zend_list_find(Z_RESVAL_PP(tmp), &type);

		zend_hash_index_find(Z_OBJPROP_P(arg), 0, (void *)&tmp);
		tc2 = (CORBA_TypeCode)zend_list_find(Z_RESVAL_PP(tmp), &type);

		CORBA_exception_init(&ev);
		if (CORBA_TypeCode_equal(tc, tc2, &ev)) {
			RETVAL_TRUE;
		} else {
			RETVAL_FALSE;
		}
		CORBA_exception_free(&ev);

		goto leave;
	} else if (zend_llist_count(props->elements_list) != 1
	           || zend_hash_num_elements(Z_OBJPROP_P(pobj)) > 0
	           || strcmp(method, "corba_typecode") != 0) {
		zend_error(E_WARNING, "Can't invoke methods on typecodes");

		RETVAL_NULL();
		goto leave;
	}

	if (zend_parse_parameters_ex(ZEND_PARSE_PARAMS_QUIET,
	                             ZEND_NUM_ARGS() TSRMLS_CC, "s",
	                             &name, &name_len) == FAILURE) {
		zend_error(E_WARNING, "CORBA_TypeCode constructor expects "
		                      "exactly one string parameter");
		goto error;
	}

	tc = phorb_find_typecode(NULL, name);
	if (!tc) {
		zend_error(E_WARNING,
		           "Could not find TypeCode for %s in "
		           "CORBA_TypeCode()", name);
		goto error;
	}

	MAKE_STD_ZVAL(handle);
	ZVAL_RESOURCE(handle, phorb_register_typecode(tc));
	zend_hash_index_update(Z_OBJPROP_P(pobj), 0, &handle,
		               sizeof(zval *), NULL);

leave:
	zval_dtor(&fname->element);
	return;

error:
	zval_dtor(pobj);
	ZVAL_NULL(pobj);

	goto leave;
}

zval phorb_get_typecode_attribute(zend_property_reference *props)
{
	zval dres;
	zval *res = &dres;
	zval **handle, *pobj = props->object;
	CORBA_TypeCode tc;
	zend_overloaded_element *aname = 
		(zend_overloaded_element *)props->elements_list->tail->data;
	char *attribute;
	int type;

	INIT_ZVAL(dres);

	convert_to_string(&aname->element);
	attribute = Z_STRVAL(aname->element);
	if (!attribute) {
		zend_error(E_WARNING, "Can't make attribute name a string");
		goto error;
	}

	zend_hash_index_find(Z_OBJPROP_P(pobj), 0, (void *)&handle);
	tc = (CORBA_TypeCode)zend_list_find(Z_RESVAL_PP(handle), &type);

	if (strcmp(attribute, "kind") == 0) {
		ZVAL_STRING(res, (char *)ORBit_tk_to_name(tc->kind), 1);
	} else if (strcmp(attribute, "flags") == 0) {
		ZVAL_LONG(res, tc->flags);
	} else if (strcmp(attribute, "length") == 0) {
		ZVAL_LONG(res, tc->length);
	} else if (strcmp(attribute, "sub_parts") == 0) {
		ZVAL_LONG(res, tc->sub_parts);
	} else if (strcmp(attribute, "subtypes") == 0) {
		int i;

		if (tc->kind != CORBA_tk_struct
		    && tc->kind != CORBA_tk_except
		    && tc->kind != CORBA_tk_union
		    && tc->kind != CORBA_tk_alias
		    && tc->kind != CORBA_tk_array
		    && tc->kind != CORBA_tk_sequence) {
			zend_error(E_WARNING, "subtypes not available for "
			                      "this type");
			goto error;
		}

		array_init(res);
		for(i = 0; i < tc->sub_parts; i++) {
			CORBA_TypeCode subtc = tc->subtypes[i];
			zval *obj, *val;

			MAKE_STD_ZVAL(obj);
			obj = phorb_init_typecode(obj);
			if (!obj) {
				zval_dtor(obj);
				FREE_ZVAL(obj);
				goto error;
			}

			MAKE_STD_ZVAL(val);
			ZVAL_RESOURCE(val, phorb_register_typecode(
				(CORBA_TypeCode)CORBA_Object_duplicate(
					(CORBA_Object)subtc, NULL)));
			zend_hash_index_update(Z_OBJPROP_P(obj), 0, &val,
			                       sizeof(zval *), NULL);

			add_next_index_zval(res, obj);
		}
	} else if (strcmp(attribute, "discriminator") == 0) {
		zval *val;

		if (tc->kind != CORBA_tk_union) {
			zend_error(E_WARNING, "discriminator not available "
			                      "for this type");
			goto error;
		}

		res = phorb_init_typecode(res);
		if (!res)
			goto error;

		MAKE_STD_ZVAL(val);
		ZVAL_RESOURCE(val, phorb_register_typecode(
			(CORBA_TypeCode)CORBA_Object_duplicate(
		        	(CORBA_Object)tc->discriminator, NULL)));
		zend_hash_index_update(Z_OBJPROP_P(res), 0, &val,
		                       sizeof(zval *), NULL);
	} else if (strcmp(attribute, "name") == 0) {
		ZVAL_STRING(res, tc->name, 1);
	} else if (strcmp(attribute, "repo_id") == 0) {
		ZVAL_STRING(res, tc->repo_id, 1);
	} else if (strcmp(attribute, "subnames") == 0) {
		int i;

		if (tc->kind != CORBA_tk_struct
		    && tc->kind != CORBA_tk_except
		    && tc->kind != CORBA_tk_union
		    && tc->kind != CORBA_tk_enum) {
			zend_error(E_WARNING, "subnames not available for "
			                      "this type");
			goto error;
		}

		array_init(res);
		for(i = 0; i < tc->sub_parts; i++)
			add_next_index_string(res, tc->subnames[i], 1);
	} else if (strcmp(attribute, "sublabels") == 0) {
		int i;

		if (tc->kind != CORBA_tk_union) {
			zend_error(E_WARNING, "sublabels not available for "
			                      "this type");
			goto error;
		}

		array_init(res);
		for(i = 0; i < tc->sub_parts; i++)
			add_next_index_long(res, tc->sublabels[i]);
	} else if (strcmp(attribute, "default_index") == 0) {
		if (tc->kind != CORBA_tk_union) {
			zend_error(E_WARNING, "default_index not available "
			                      "for this type");
			goto error;
		}

		ZVAL_LONG(res, tc->default_index);
	} else if (strcmp(attribute, "recurse_depth") == 0) {
		if (tc->kind != CORBA_tk_sequence) {
			zend_error(E_WARNING, "recurse_depth not available "
			                      "for this type");
			goto error;
		}

		ZVAL_LONG(res, tc->recurse_depth);
	} else if (strcmp(attribute, "digits") == 0) {
		if (tc->kind != CORBA_tk_fixed) {
			zend_error(E_WARNING, "digits not available for"
			                      "this type");
			goto error;
		}

		ZVAL_LONG(res, tc->digits);
	} else if (strcmp(attribute, "scale") == 0) {
		if (tc->kind != CORBA_tk_fixed) {
			zend_error(E_WARNING, "scale not available for "
			                      "this type");
			goto error;
		}

		ZVAL_LONG(res, tc->scale);
	} else if (strcmp(attribute, "_hash") == 0) {
		CORBA_Environment ev;

		CORBA_exception_init(&ev);
		ZVAL_LONG(res, CORBA_Object_hash((CORBA_Object)tc,
		                                 LONG_MAX, &ev));
		CORBA_exception_free(&ev);
	} else {
		zend_error(E_WARNING, "Non-existing attribute `%s'",
		           attribute);
	}

leave:
	zval_dtor(&aname->element);
	return dres;

error:
	zval_dtor(res);
	ZVAL_BOOL(res, 0);

	goto leave;
}

int phorb_set_typecode_attribute(zend_property_reference *props, zval *value)
{
	zend_overloaded_element *aname = 
		(zend_overloaded_element *)props->elements_list->tail->data;

	zend_error(E_WARNING, "Typecode attributes are readonly");

	zval_dtor(&aname->element);
	return FAILURE;
}

void phorb_serialize_typecode_sleep(INTERNAL_FUNCTION_PARAMETERS)
{
	CORBA_TypeCode tc;
	zval **handle;
	char *repoid;
	int type;

	zend_hash_index_find(Z_OBJPROP_P(this_ptr), 0, (void *)&handle);
	tc = (CORBA_TypeCode)zend_list_find(Z_RESVAL_PP(handle), &type);
	array_init(return_value);
	repoid = (char *)tc->repo_id;

	if (!repoid || !repoid[0])
		repoid = "<unknown>";

	add_property_string(this_ptr, "_repoid", repoid, 1);
	add_next_index_string(return_value, "_repoid", 1);
}

void phorb_serialize_typecode_wakeup(INTERNAL_FUNCTION_PARAMETERS)
{
	CORBA_TypeCode tc;
	zval *handle, **repoid;
	int type;

	zend_hash_find(Z_OBJPROP_P(this_ptr), "_repoid", 8, (void *)&repoid);
	if (!repoid) {
		zend_error(E_ERROR, "Bad serialization of CORBA_TypeCode "
		                    "object, no repoid found");
		RETURN_FALSE;
	}
	if (Z_TYPE_PP(repoid) != IS_STRING) {
		zend_error(E_ERROR, "Bad serialization of CORBA_TypeCode "
		                    "object, repoid not a string");
		RETURN_FALSE;
	}

	tc = phorb_find_typecode(NULL, Z_STRVAL_PP(repoid));
	if (!tc) {
		zend_error(E_WARNING, "Could not find repoid `%s' in "
		                      "typecodes", Z_STRVAL_PP(repoid));
		RETURN_FALSE;
	}
	zval_ptr_dtor(repoid);
	zend_hash_del(Z_OBJPROP_P(this_ptr), "_repoid", 8);

	MAKE_STD_ZVAL(handle);
	ZVAL_RESOURCE(handle, phorb_register_typecode(tc));

	zend_hash_index_update(Z_OBJPROP_P(this_ptr), 0, &handle,
	                       sizeof(zval *), NULL);

	RETURN_TRUE;
}

CORBA_long phorb_enum_find_member(CORBA_TypeCode tc, zval *val)
{
	char *str;
	CORBA_long i;

	convert_to_string(val);
	str = Z_STRVAL_P(val);
	if (!str)
		return -1;

	for(i = 0; i < tc->sub_parts; i++)
		if (strcmp(tc->subnames[i], str) == 0)
			return i;

	return -1;
}

#define ZVAL_RES_FIND_MEMBER(kind, type) \
	case CORBA_tk_##kind: do { \
		CORBA_##type val = Z_RESVAL_P(discriminator); \
		\
		for(i = 0; i < tc->sub_parts; i++) { \
			CORBA_##type label = \
				*(CORBA_##type *)tc->sublabels[i]; \
			\
			if (label == val) \
				return i; \
		} \
		break; \
	} while(0)

CORBA_long phorb_union_find_arm(CORBA_TypeCode tc, zval *discriminator)
{
	CORBA_unsigned_long i;

	/* assuming discriminator IS_RESOURCE */

	switch(tc->discriminator->kind) {
	ZVAL_RES_FIND_MEMBER(char, char);
	ZVAL_RES_FIND_MEMBER(octet, octet);
	ZVAL_RES_FIND_MEMBER(short, short);
	ZVAL_RES_FIND_MEMBER(ushort, unsigned_short);
	ZVAL_RES_FIND_MEMBER(long, long);
	ZVAL_RES_FIND_MEMBER(ulong, unsigned_long);
	ZVAL_RES_FIND_MEMBER(longlong, long_long);
	ZVAL_RES_FIND_MEMBER(ulonglong, unsigned_long_long);
	case CORBA_tk_enum: {
		CORBA_unsigned_long val =
			phorb_enum_find_member(
				tc->discriminator, discriminator);

		for(i = 0; i < tc->sub_parts; i++)
			if (tc->sublabels[i] == val)
				return i;
		break;
	}
	case CORBA_tk_boolean: {
		CORBA_boolean val = Z_RESVAL_P(discriminator);

		for(i = 0; i < tc->sub_parts; i++)
			if (!tc->sublabels[i] == !val)
				return i;
		break;
	}
	default:
		zend_error(E_WARNING,
		           "Unsupported discriminator type %d",
		           tc->discriminator->kind);
	}

	return (tc->default_index >= 0) ? tc->default_index : -1;
}

#undef ZVAL_RES_FIND_MEMBER
